<?php
/*
Copyright (C) 2018 Daniel Wirths

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

namespace src\server;

use src\log\logRepository;
use src\log\logService;
use src\server\serverRepository;
use src\settings\settingsRepository;
use src\tracker\trackerDataRepository;
use src\tracker\trackerDeviceRepository;
use src\tracker\trackerModelRepository;
use src\tracker\trackerService;

class serverController
{
	private $clientConnected;
	private $dataInputSave;
	private $lastCheckTimeoutSendToClient;
	private $match;
	private $serverStop;
	private $settings;
	private $socket;
	private $trackerModel;
	private $trackerModelKey;
	private $waitTimeoutSendToClient;

 	public function __construct(
		logRepository $logRepository,
		logService $logService,
		serverRepository $serverRepository,
		settingsRepository $settingsRepository,
		trackerDataRepository $trackerDataRepository,
		trackerDeviceRepository $trackerDeviceRepository,
		trackerModelRepository $trackerModelRepository,
		trackerService $trackerService
	)
	{
		$this->logRepository = $logRepository;
		$this->logService = $logService;
		$this->serverRepository = $serverRepository;
		$this->settingsRepository = $settingsRepository;
		$this->trackerDataRepository = $trackerDataRepository;
		$this->trackerDeviceRepository = $trackerDeviceRepository;
		$this->trackerModelRepository = $trackerModelRepository;
		$this->trackerService = $trackerService;

		declare(ticks = 1);

		pcntl_signal(SIGTERM, function () { $this->closeOnExit(); });
		pcntl_signal(SIGINT, function () { $this->closeOnExit(); });
	}

	public function init()
	{
		$this->logService->logFile("Server gestartet.", 'LOG');

		$this->settings = $this->settingsRepository->all();
		$this->trackerModelAll = $this->trackerModelRepository->all();

		if(!empty($this->settings[0]["wsport"]))
		{
			$this->makeSocket('TCP', $this->settings[0]["wsport"]);

			$this->clients["TCP_".$this->settings[0]["wsport"]] = $this->socket["TCP_".$this->settings[0]["wsport"]];
		}

		if(!empty($this->trackerModelAll))
		{
			foreach($this->trackerModelAll as $value)
			{
				$this->makeSocket($value['protocol'], $value['port']);

				$this->clients[$value['protocol']."_".$value['port']] = $this->socket[$value['protocol']."_".$value['port']];
			}
		}
		else
		{
			$this->logService->logFile("Keine Ports gefunden", 'ERROR');
			die();
		}

		$this->waitForConnection();
	}

	protected function closeOnExit()
	{
		foreach($this->socket as $key=>$socketClose)
		{
			socket_shutdown($socketClose, 2);
			socket_close($socketClose);
			unset($this->socket[$key]);
		}
// Hier vielleicht die PID in Datei?? zum Check if Run
		$this->logService->logFile("Server heruntergefahren.", 'WARNING');
		exit;
	}

	protected function makeSocket($protocol, $port)
	{
		$this->createNewSocket($protocol, $port);
		$this->bindSocket($protocol, $port);
		$this->listenSocket($protocol, $port);
	}

	protected function createNewSocket($protocol, $port)
	{
		if($protocol === 'UDP')
		{
			$this->socket[$protocol."_".$port] = @socket_create(AF_INET, SOCK_DGRAM, SOL_UDP);
		}

		if($protocol === 'TCP')
		{
			$this->socket[$protocol."_".$port] = @socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
		}

		socket_set_option($this->socket[$protocol."_".$port], SOL_SOCKET, SO_REUSEADDR, 1);
		socket_set_option($this->socket[$protocol."_".$port], SOL_SOCKET, SO_KEEPALIVE, 1);

		if($this->socket[$protocol."_".$port] === FALSE)
		{
			$errorcode = socket_last_error();
			$errormsg = socket_strerror($errorcode);

			$this->logService->logFile("Socket konnte nicht erstellt werden. Error: [$errorcode] $errormsg", 'ERROR');
			die();
		}
		else
		{
			$this->logService->logFile("Socket erfolgreich erstellt.", 'LOG');
		}
	}

	protected function bindSocket($protocol, $port)
	{
		if(@socket_bind($this->socket[$protocol."_".$port], $this->settings[0]->ip, $port) === FALSE)
		{
			$errorcode = socket_last_error();
			$errormsg = socket_strerror($errorcode);

			$this->logService->logFile("Socket konnte nicht gebunden werden. Error: [$errorcode] $errormsg", 'ERROR');
			die();
		}
		else
		{
			$this->logService->logFile("Socket-Port ".$port." gebunden.", 'LOG');
		}
	}

	protected function listenSocket($protocol, $port)
	{
		if(@socket_listen($this->socket[$protocol."_".$port], $this->settings[0]->maxClientsWait) === FALSE)
		{
			$errorcode = socket_last_error();
			$errormsg = socket_strerror($errorcode);

			$this->logService->logFile("Socket konnte nicht abgehört werden. Error: [$errorcode] $errormsg", 'ERROR');
			die();
		}
		else
		{
			$this->logService->logFile("Socket wird abgehört.", 'LOG');
		}
	}

	protected function waitForConnection()
	{
		while(true)
		{
			$this->serverRepository->updateConnection();

			$read = $this->clients;

			if(@socket_select($read, $write, $except, 0) === false)
			{
				$errorcode = socket_last_error();
				$errormsg = socket_strerror($errorcode);

				$this->logService->logFile("Socket konnte nicht gelesen werden.\nError: [$errorcode] $errormsg", 'ERROR');
				die();
			}
			else
			{
				if(isset($this->socket[key($read)]) && in_array($this->socket[key($read)], $read))
				{
					$client = socket_accept($this->socket[key($read)]);
					$this->clients[] = $client;

					socket_getpeername($client, $adress, $port);
					socket_getsockname($client, $adressSocket, $portSocket);

					$this->logService->logFile($adress . ":" . $port . " ist nun an Port " . $portSocket . " verbunden.", 'LOG');

					$this->logService->logFile("Es sind ".(count($this->clients) - count($this->socket))." Client(s) verbunden.", 'LOG');

					$clientKeyReaded = array_search($this->socket[key($read)], $read);
					unset($read[$clientKeyReaded]);
				}

				foreach($read as $clientReading)
				{
					$dataInput = '';
					$buff = '';

					while(@socket_recv($clientReading, $buff, 1024, MSG_DONTWAIT) != 0)
					{
						$dataInput .= $buff;
					}

					$dataInput = trim($dataInput);
					$clientKey = array_search($clientReading, $this->clients);

					if($dataInput == false or $dataInput == '')
					{
						$this->disconnect($clientKey);
					}
					else
					{
						socket_getsockname($clientReading, $adress, $port);

						if($port == $this->settings[0]["wsport"])
						{
							if(empty($this->clientConnected[$clientKey]['handshake']))
							{
								$this->clientConnected[$clientKey]['handshake'] = $this->handshake($clientKey, $dataInput);

								if($this->clientConnected[$clientKey]['handshake'] == true)
								{
									$this->logService->logFile("Handshake erfolgreich.", 'LOG');
								}
								else
								{
									$this->disconnect($clientKey);
									continue;
								}

								if($this->checkSession($clientKey) == false)
								{
									continue;
								}
							}
						}
						else
						{
							$this->trackerModelKey = array_search($port, array_column($this->trackerModelAll, 'port'));

							preg_match_all('/'.$this->trackerModelAll[$this->trackerModelKey]["pattern"].'/', $dataInput, $matches, PREG_SET_ORDER);

							if($matches)
							{
								foreach($matches as $this->match)
								{
									$this->findTrackerByImei();

									if(!empty($this->trackerDevice))
									{
										$this->clientConnected[$clientKey]['trackerId'] = $this->trackerDevice["id"];
										$this->clientConnected[$clientKey]['handshakeSend'] = true;

										$this->prepareDataForInsert();

										$id = $this->trackerDataRepository->insertTrackerData(
											$this->dataInputSave["trackerDeviceId"],
											$this->dataInputSave["imei"],
											$this->dataInputSave["time"],
											$this->dataInputSave["status"],
											$this->dataInputSave["latitude"],
											$this->dataInputSave["longitude"],
											$this->dataInputSave["speed"],
											$this->dataInputSave["direction"],
											$this->dataInputSave["event"],
											$this->dataInputSave["voltage"],
											$this->dataInputSave["voltagePercent"]
										);

										if(!empty($id))
										{
											$this->logService->logFile("Daten gespeichert.", 'LOG');
										}
										else
										{
											$this->logService->logFile("Eintragung in Datenbank fehlgeschlagen.", 'ERROR');
										}
									}
								}
							}
							else
							{
								$this->logService->logFile("Eintragung fehlgeschlagen: ".$dataInput, 'NOTICE');
							}
						}
					}
				}
			}

			if(empty($this->waitTimeoutSendToClient))
			{
				$this->waitTimeoutSendToClient = 0.9; // 0.5 = halbe Sekunde, 1 = eine Sekunde
			}

			if($this->lastCheckTimeoutSendToClient < microtime(true))
			{
				$this->lastCheckTimeoutSendToClient = $this->waitTimeoutSendToClient + microtime(true);

				foreach($this->clients as $clientKey=>$sendToClient)
				{
					if(in_array($sendToClient, $this->socket))
					{
						continue;
					}

					socket_getsockname($sendToClient, $adress, $port);

					if($port !== $this->settings[0]["wsport"])
					{


						continue;
					}

					if(empty($this->clientConnected[$clientKey]['handshake']) or $this->clientConnected[$clientKey]['handshake'] !== true)
					{
						continue;
					}

					if($this->checkSession($clientKey) == false)
					{
						continue;
					}

					if($this->clientConnected[$clientKey]['path'] == '/dashboard')
					{
						if(empty($this->clientConnected[$clientKey]['trackerDeviceId']))
						{
							$this->clientConnected[$clientKey]['trackerDeviceId'] = $this->clientConnected[$clientKey]['session']['trackerDeviceId'];
						}

						if(empty($this->clientConnected[$clientKey]['lastTrackerDataId']))
						{
							$lastId = $this->trackerDataRepository->getLastIdByTrackerDeviceId($this->clientConnected[$clientKey]['trackerDeviceId']);
							$this->clientConnected[$clientKey]['lastTrackerDataId'] = $lastId['id'];
						}

						$newTrackerData = $this->trackerDataRepository->getLastTrackingPointsById(
							$this->clientConnected[$clientKey]['trackerDeviceId'],
							$this->clientConnected[$clientKey]['lastTrackerDataId']
						);

						if(in_array($this->clientConnected[$clientKey]['trackerDeviceId'], array_column($this->clientConnected, 'trackerId')))
						{
							if(empty($this->clientConnected[$clientKey]['sendTrackerStatus']))
							{
								$this->clientConnected[$clientKey]['sendTrackerStatus'] = true;

								$dataToSend  = 'trackerId: '.$this->clientConnected[$clientKey]['trackerDeviceId']."\n";
								$dataToSend .= 'trackerStatus: online'."\n";

								socket_write($sendToClient, $this->encode($dataToSend));
							}
						}
						else
						{
							if(!empty($this->clientConnected[$clientKey]['sendTrackerStatus']))
							{
								$this->clientConnected[$clientKey]['sendTrackerStatus'] = false;

								$dataToSend  = 'trackerId: '.$this->clientConnected[$clientKey]['trackerDeviceId']."\n";
								$dataToSend .= 'trackerStatus: offline'."\n";

								socket_write($sendToClient, $this->encode($dataToSend));
							}
						}

						if(empty($newTrackerData))
						{
							continue;
						}

						$this->clientConnected[$clientKey]['lastTrackerDataId'] = end($newTrackerData)['id'];

						foreach($newTrackerData as $data)
						{
							$dataToSend  = 'id: '.$data['id']."\n";
							$dataToSend .= 'trackerDeviceId: '.$data['trackerDeviceId']."\n";
							$dataToSend .= 'recivedTime: '.date("d.m.Y - H:i:s", $data['recivedTime'])."\n";
							$dataToSend .= 'imei: '.$data['imei']."\n";
							$dataToSend .= 'time: '.$data['time']."\n";
							$dataToSend .= 'status: '.$data['status']."\n";
							$dataToSend .= 'latitude: '.$data['latitude']."\n";
							$dataToSend .= 'longitude: '.$data['longitude']."\n";
							$dataToSend .= 'speed: '.$data['speed']."\n";
							$dataToSend .= 'direction: '.$data['direction']."\n";
							$dataToSend .= 'event: '.$data['event']."\n";
							$dataToSend .= 'voltage: '.$data['voltage']."\n";
							$dataToSend .= 'voltagePercent: '.$data['voltagePercent']."\n";

							socket_write($sendToClient, $this->encode(e($dataToSend)));
						}
					}
					elseif($this->clientConnected[$clientKey]['path'] == '/log')
					{
						if(empty($this->clientConnected[$clientKey]['lastLogId']))
						{
							$lastId = $this->logRepository->getLastId();
							$this->clientConnected[$clientKey]['lastLogId'] = $lastId['id'];
						}

						$newLogs = $this->logRepository->getLogsByLastId($this->clientConnected[$clientKey]['lastLogId']);

						if(empty($newLogs))
						{
							continue;
						}

						$this->clientConnected[$clientKey]['lastLogId'] = end($newLogs)['id'];

						foreach($newLogs as $log)
						{
							socket_write($sendToClient, $this->encode($this->logService->readLog($log["time"], $log["logData"], $log["level"])));
						}
					}
				}
			}

			// $decodeContent = $this->decode($dataInput);
		}
	}

	protected function findTrackerByImei()
	{
		$imeiPosition = $this->trackerModelAll[$this->trackerModelKey]["imeiPosition"];

		$this->trackerDevice = $this->trackerDeviceRepository->findByImei($this->match[$imeiPosition]);
	}

	protected function prepareDataForInsert()
	{
		$method = $this->trackerModelAll[$this->trackerModelKey]["timeFunction"];

		$this->dataInputSave["trackerDeviceId"] = $this->trackerDevice["id"];
		$this->dataInputSave["imei"] = $this->trackerDevice["imei"];
		$this->dataInputSave["time"] = $this->trackerService->$method(
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["timeDayPosition"]],
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["timeMonthPosition"]],
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["timeYearPosition"]],
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["timeHourPosition"]],
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["timeMinutePosition"]],
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["timeSecondPosition"]]
		);
		$this->dataInputSave["status"] = $this->match[$this->trackerModelAll[$this->trackerModelKey]["statusPosition"]];

		$method = $this->trackerModelAll[$this->trackerModelKey]["coordinatesBFunction"];

		$this->dataInputSave["latitude"] = $this->trackerService->$method(
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["coordinatesBPosition"]],
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["coordinatesNSPosition"]]
		);

		$method = $this->trackerModelAll[$this->trackerModelKey]["coordinatesLFunction"];

		$this->dataInputSave["longitude"] = $this->trackerService->$method(
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["coordinatesLPosition"]],
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["coordinatesEWPosition"]]
		);

		$method = $this->trackerModelAll[$this->trackerModelKey]["speedFunction"];

		$this->dataInputSave["speed"] = $this->trackerService->$method(
			$this->match[$this->trackerModelAll[$this->trackerModelKey]["speedPosition"]]
		);

		$this->dataInputSave["direction"] = $this->match[$this->trackerModelAll[$this->trackerModelKey]["directionPosition"]];
		$this->dataInputSave["event"] = $this->match[$this->trackerModelAll[$this->trackerModelKey]["eventPosition"]];
		$this->dataInputSave["voltage"] = $this->match[$this->trackerModelAll[$this->trackerModelKey]["voltagePosition"]];
		$this->dataInputSave["voltagePercent"] = $this->match[$this->trackerModelAll[$this->trackerModelKey]["voltagePercentPosition"]];
	}

	private function handshake($clientKey, $headers)
	{
		if(preg_match("/Sec-WebSocket-Version: (.*)\r\n/", $headers, $match))
		{
			$version = $match[1];
		}
		else
		{
			$this->logService->logFile("Client unterstützt das WebSocket-Protokoll nicht.", 'ERROR');

			return false;
		}

		if($version == 13)
		{
			if(preg_match("/Cookie: (.*)\r\n/", $headers, $match))
			{
				preg_match("/".session_name()."=(.*)(| |;)/",  $match[1], $matchCookie);

				$sessionId = $matchCookie[1];
			}

			if(preg_match("/GET (.*) HTTP/", $headers, $match))
			{
				$path = $match[1];
			}

			if(preg_match("/Sec-WebSocket-Key: (.*)\r\n/", $headers, $match))
			{
				$websocketKey = $match[1];

				$acceptKey = $websocketKey . '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
				$acceptKey = base64_encode(sha1($acceptKey, true));

				$handshake = "HTTP/1.1 101 Switching Protocols\r\n" .
					"Upgrade: websocket\r\n" .
					"Connection: Upgrade\r\n" .
					"Sec-WebSocket-Accept: $acceptKey" .
					"\r\n\r\n";

				socket_write($this->clients[$clientKey], $handshake);

				$this->clientConnected[$clientKey]['sessionId'] = $sessionId;
				$this->clientConnected[$clientKey]['socket'] = $clientKey;
				$this->clientConnected[$clientKey]['path'] = $path;

				return true;
			}
		}
	}

	private function getSession($sessionId)
	{
		if(file_exists(session_save_path() . '/sess_' . $sessionId))
		{
			$sessionDataFile = file_get_contents(session_save_path() . '/sess_' . $sessionId);

			@session_start([
				'use_cookies' => 0,
				'use_only_cookies' => 0,
				'cache_limiter' => 0]
			);

			session_decode($sessionDataFile);

			$sessionData = $_SESSION;

			unset($_SESSION);
			session_destroy();

			return $sessionData;
		}
	}

	private function checkSession($clientKey)
	{
		$this->clientConnected[$clientKey]['session'] = $this->getSession($this->clientConnected[$clientKey]['sessionId']);

		//Check Login Time
		if(isset($this->clientConnected[$clientKey]['session']['loginUID']))
		{
			return true;
		}
		else
		{
			$dataToSend  = 'disconnect: true'."\n";

			socket_write($this->clients[$clientKey], $this->encode($dataToSend));

			$this->disconnect($clientKey);

			return false;
		}
	}

	private function decode($input)
	{
		$len = ord($input[1]) & 127;

		if($len === 126)
		{
			$ofs = 8;
		}
		elseif($len === 127)
		{
			$ofs = 14;
		}
		else
		{
			$ofs = 6;
		}

		$text = '';

		for($i = $ofs; $i < strlen($input); $i++)
		{
			$text .= $input[$i] ^ $input[$ofs - 4 + ($i - $ofs) % 4];
		}

		return $text;
	}

	private function encode($text)
	{
		$b = 129;

		$len = strlen($text);

		if($len < 126)
		{
			return pack('CC', $b, $len) . $text;
		}
		elseif($len < 65536)
		{
			return pack('CCn', $b, 126, $len) . $text;
		}
		else
		{
			return pack('CCNN', $b, 127, 0, $len) . $text;
		}
	}

	private function disconnect($clientKey)
	{
		@socket_getpeername($this->clients[$clientKey], $adress, $port);

		socket_close($this->clients[$clientKey]);

		unset($this->clients[$clientKey]);
		unset($this->clientConnected[$clientKey]);

		$this->logService->logFile($adress . ":" . $port . " Client getrennt. " . (count($this->clients) - count($this->socket)) . " Client(s) verbunden.", 'LOG');
	}
}
?>